# -*- coding: utf-8 -*-

from __future__ import absolute_import

import numpy

from rasterio.mask import mask


def summary(raster, geometry=None, all_touched=True, mean_only=False,
            bounds=None):
    """Return ``ST_SummaryStats`` style stats for the given raster.

    If ``geometry`` is provided, we mask the raster with the given geometry and
    return the stats for the intersection.

    If ``all_touched`` is set, we include every pixel that is touched by the
    given geometry. If set to ``False``, we only include pixels that are
    "mostly" inside the given geometry (the calculation is done by Rasterio).

    If ``mean_only`` is ``True`` we only return the mean value of the pixels,
    not the full set of stats.

    If ``bounds`` is passed, it should be a two-tuple of (min, max) to use for
    filtering raster pixels. If not provided, we exclude anything equal to the
    raster no data value.

    If ``mean_only`` is ``False``, the values return correspond to::

        ['count', 'data_count', 'sum', 'mean', 'min', 'max', 'std']

    The difference betwee ``count`` and ``data_count`` is that ``data_count``
    is the count of all pixels that are either within ``bounds`` or not equal
    to the raster no data value.

    If ``mean_only`` is ``True``, we simply return a ``float`` or ``None``
    representing the mean value of the matching pixels.

    """
    def no_result(mean_only):
        if mean_only:
            return None
        else:
            return [0, None, None, None, None, None, None]

    try:
        if geometry:
            result, _ = mask(
                raster, [geometry], crop=True, all_touched=all_touched,
            )
            pixels = result.data.flatten()
        else:
            pixels = raster.read(1).flatten()
    except:
        return no_result(mean_only)

    if bounds:
        score_mask = numpy.logical_and(
            numpy.greater_equal(pixels, bounds[0]),
            numpy.less_equal(pixels, bounds[1]),
        )
    else:
        score_mask = numpy.not_equal(pixels, raster.nodata),

    scored_pixels = numpy.extract(score_mask, pixels)
    if len(scored_pixels):
        if mean_only:
            return round(scored_pixels.mean(), 3)
        else:
            return map(lambda x: round(x, 3), [
                len(pixels),
                len(scored_pixels),
                scored_pixels.sum(),
                scored_pixels.mean(),
                scored_pixels.std(),
                scored_pixels.min(),
                scored_pixels.max(),
            ])
    else:
        no_result(mean_only)



# -*- coding: utf-8 -*-

from __future__ import absolute_import

import click
import rasterio
import tablib

from functools import partial
from tabulate import tabulate

from shapely.wkt import loads as load_wkt
from shapely.geometry import mapping, Point

from winston.utils import buffer_point
from winston.stats import summary


@click.command()
@click.argument('fnames', nargs=-1)
@click.option('--not-all-touched', is_flag=True,
              help='Only include pixels that have centroids inside the area. '
                   'Default is to include all pixels touched by the area.')
@click.option('--proportional', 'radius_mode', flag_value='prop',
              help='Use a radius proportional to latitude')
@click.option('--fixed', 'radius_mode', flag_value='fixed', default=True,
              help='Always use a buffer radius of 25/111.13 degrees (default)')
@click.option('--radius', '-r', default=25.0,
              help='Point radius in km (pass zero to disable point buffering)')
@click.option('--bounds', nargs=2, type=float,
              help='Optional bounds for clipping values rather than using '
                   'the raster no data value')
@click.option('--mean-only', '-o', is_flag=True,
              help='Only return scores, not all summary stats')
@click.option('--meta', '-m', is_flag=True,
              help='Include metadata about the raster in the output')
@click.option('--wkt', '-w', multiple=True,
              help='WKT string (can be used multiple times)')
@click.option('--wkt-file', '-W',
              help='Text file containing WKT strings (one per line)')
@click.option('--csv-file', '-c',
              help='CSV file containing location data')
def main(fnames=None, not_all_touched=False, bounds=None, radius=None,
         radius_mode=None, mean_only=False, meta=False, wkt=None,
         wkt_file=None, csv_file=None):
    all_touched = not not_all_touched

    if not len(fnames):
        click.echo('No file names given. Exiting.')
        return

    geometries = []
    # First, parse any WKT strings passed in
    if wkt:
        geometries.extend([load_wkt(s) for s in wkt])

    # Next, read any WKT files (plain text, one WKT per line)
    if wkt_file:
        geometries.extend([load_wkt(l) for l in open(wkt_file).readlines()])

    # CSV must have either 'lat' and 'lon'/'lng' columns *or* a 'geom'/'wkt'
    # column (if you want to include WKT for a geometry rather than
    # coordinates. We check for coords first, then geometries.
    if csv_file:
        d = tablib.Dataset()
        d.csv = open(csv_file).read()
        for record in d.dict:
            lon = record.get('lon', record.get('lng'))
            lat = record.get('lat')
            geom = record.get('geom', record.get('wkt'))
            if lon and lat:
                geometries.append(Point(float(lon), float(lat)))
            elif geom:
                geometries.append(load_wkt(geom))

    # Take any points and buffer them if radius is > 0.
    source_geometries = geometries
    if radius > 0:
        buf = partial(buffer_point, radius=radius, radius_mode=radius_mode)
        geometries = map(buf, geometries)

    # Turn all the geometry objects into GeoJSON for rasterio
    geometries = map(mapping, geometries)

    if bounds:
        click.echo(
            ' - reporting stats for pixels where {} <= x <= {}'.format(*bounds)
        )
    else:
        click.echo(
            ' - reporting stats for all pixels with values'
        )

    if not len(geometries):
        click.echo(
            ' - no geometries provided, reporting stats for the entire raster'
        )
        geometries = [None]
        source_geometries = ['Entire raster']
    else:
        if radius == 0:
            click.echo(' - points not buffered')
        else:
            if radius_mode == 'prop':
                click.echo(
                    ' - buffers ({}km) are proportional to latitude'.format(
                        radius,
                    )
                )
            else:
                click.echo(
                    ' - buffers fixed at a {}/111.13 degree radius'.format(
                        radius,
                    )
                )

        if all_touched:
            click.echo(
                ' - pixels counted if any part is inside the geometry'
            )
        else:
            click.echo(
                ' - pixels only counted if the centroid is inside the geometry'
            )

    rasters = []
    for fname in fnames:
        try:
            rasters.append(rasterio.open(fname))
        except Exception as ex:
            click.echo('Error loading {} ({}). Aborting.'.format(fname, ex))

    for raster in rasters:
        results = []

        # We keep the source geometries so we can return them with the results
        # rather than the buffered versions. We also pass the source to Dekker
        # because it does its own buffering.
        for src, geometry in zip(source_geometries, geometries):
            name = str(src)
            if len(name) > 28:
                name = name[:25] + '...'

            row = [name]

            stats = summary(raster, geometry, all_touched, mean_only, bounds)
            if mean_only:
                row.append(stats)
            else:
                row.extend(stats)

            results.append(row)

        headers = [
            'geometry',
        ]
        if not mean_only:
            headers.extend(
                ['count', 'scored', 'sum', 'mean', 'min', 'max', 'std']
            )
        else:
            headers.append('mean')

        click.echo()
        click.echo(raster.name)
        if meta:
            click.echo(tabulate(
                [
                    ('CRS', str(raster.crs)),
                    ('Bounds', raster.bounds),
                    ('Size', '{} x {}'.format(raster.width, raster.height)),
                    ('Tiled', raster.is_tiled),
                    ('Block size', '{} x {}'.format(*raster.block_shapes[0])),
                    ('Driver', raster.driver),
                    ('Data type', ', '.join(raster.dtypes)),
                    (
                        'Compression',
                        raster.compression.value if raster.compression else ''
                    ),
                ],
                ['Meta', 'Data'],
                tablefmt='psql',
            ))
        click.echo(tabulate(results, headers, tablefmt='psql'))


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from setuptools import setup

with open('README.rst') as readme_file:
    readme = readme_file.read()

with open('HISTORY.rst') as history_file:
    # Skipping the first line reduces the heading level by 1
    history = ''.join(history_file.readlines()[1:])

requirements = [
    'click==6.6',
    'rasterio==0.36.0',
    'requests[security]==2.4.3',
    'shapely==1.5.17',
    'numpy==1.11.2',
    'tablib==0.11.2',
]

setup(
    name='winston',
    version='0.1.1',
    description="A bit like PostGIS ST_SummaryStats for files on disk",
    long_description=readme + '\n\n' + history,
    author="James Rutherford",
    author_email='james.rutherford@maplecroft.com',
    url='https://github.com/Maplecroft/Winston',
    packages=[
        'winston',
    ],
    package_dir={'winston': 'winston'},
    entry_points={
        'console_scripts': [
            'winston=winston.cli:main'
        ],
    },
    include_package_data=True,
    install_requires=requirements,
    license="MIT license",
    zip_safe=False,
    keywords='winston',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        "Programming Language :: Python :: 2",
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
    ],
)



# -*- coding: utf-8 -*-

from __future__ import absolute_import

import math

from shapely.geometry import Point


def km_to_degrees(km, latitude=None):
    if not latitude:
        return km / 111.13
    else:
        earth_radius = 6371.0
        degrees_to_radians = math.pi / 180.0
        radians_to_degrees = 180.0 / math.pi
        r = earth_radius * math.cos(latitude * degrees_to_radians)
        return (km / r) * radians_to_degrees


def buffer_point(geometry, radius, radius_mode):
    if isinstance(geometry, Point) and radius > 0:
        if radius_mode == 'fixed':
            return geometry.buffer(km_to_degrees(radius))
        else:
            return geometry.buffer(
                km_to_degrees(radius, geometry.coords[0][1])
            )
    else:
        return geometry
